// Hyperbolic Rogue
// Copyright (C) 2011 Zeno Rogue, see 'hyper.cpp' for details

// cells the game is played on

int fix6(int a) { return (a+96)% 6; }

// information for the game
struct gcell {
  bool iswall;
  int item, monst;
  int cpdist, mpdist, pathdist;
  ld heat;
  };

struct cell : gcell {
  int type; // 6 for hexagons, 7 for heptagons
  heptagon *master;
  cell *mov[7]; // meaning very similar to heptagon::move
  int spn[7];
  };

int cellcount = 0;

void initcell(cell *c); // from game.cpp

cell *newCell(int type, heptagon *master) {
  cell *c = new cell;
  cellcount++;
  c->type = type;
  c->master = master;
  for(int i=0; i<7; i++) c->mov[i] = NULL;
  initcell(c);
  return c;
  }

// very similar to createMove in heptagon.cpp
void createMov(cell *c, int d) {
  if(c->mov[d]) return;
  else if(c->type == 7) {
    cell *n = newCell(6, c->master);

    c->mov[d] = n; n->mov[0] = c;
    c->spn[d] = 0; n->spn[0] = d;
    
    heptspin hs; hs.h = c->master; hs.spin = d;
    
    heptspin hs2 = hsstep(hsspin(hs, 3), 3);
    
    hs2.h->c7->mov[hs2.spin] = n; n->mov[2] = hs2.h->c7;
    hs2.h->c7->spn[hs2.spin] = 2; n->spn[2] = hs2.spin;
    
    hs2 = hsstep(hsspin(hs, 4), 4);
    hs2.h->c7->mov[hs2.spin] = n; n->mov[4] = hs2.h->c7;
    hs2.h->c7->spn[hs2.spin] = 4; n->spn[4] = hs2.spin;
    }
  
  else if(d == 5) {
    createMov(c->mov[0], fixrot(c->spn[0]+1));
    
    c->mov[5] = c->mov[0]->mov[fixrot(c->spn[0]+1)]; 
    c->spn[5] = fix6(c->mov[0]->spn[fixrot(c->spn[0]+1)] + 1);
    }
  
  else if(d == 1) {
    createMov(c->mov[0], fixrot(c->spn[0]-1));
    
    c->mov[1] = c->mov[0]->mov[fixrot(c->spn[0]-1)]; 
    c->spn[1] = fix6(c->mov[0]->spn[fixrot(c->spn[0]-1)] - 1);
    }
  
  else if(d == 3) {
    createMov(c->mov[2], fixrot(c->spn[2]-1));
    c->mov[3] = c->mov[2]->mov[fixrot(c->spn[2]-1)];
    c->spn[3] = fix6(c->mov[2]->spn[fixrot(c->spn[2]-1)] - 1);
    }
  }

// similar to heptspin from heptagon.cpp
struct cellwalker {
  cell *c;
  int spin;
  };

void cwspin(cellwalker& cw, int d) {
  cw.spin = (cw.spin+d + 42) % cw.c->type;
  }

void cwstep(cellwalker& cw) {
  createMov(cw.c, cw.spin);
  int nspin = cw.c->spn[cw.spin];
  cw.c = cw.c->mov[cw.spin];
  cw.spin = nspin;
  }

// approximate distance of c from origin
int celldist(cell *c) {
  if(c->type == 7) return c->master->distance;
  int d1 = c->mov[0]->master->distance;
  int d2 = c->mov[2]->master->distance;
  int d3 = c->mov[4]->master->distance;
  if((d1+d2+d3) % 3 == 2) return ((d1+d2+d3) / 3);
  else return ((d1+d2+d3) / 3);
  }

// initializer (also inits origin from heptagon.cpp)
void initcells() {
  origin.s = hsOrigin;
  for(int i=0; i<7; i++) origin.move[i] = NULL;
  origin.distance = 0;
  origin.c7 = newCell(7, &origin);
  }

