// Hyperbolic Rogue
// Copyright (C) 2011 Zeno Rogue, see 'hyper.cpp' for details

// game state
int gold = 0, kills[4], explore[10];
bool playerdead = false;
bool playermoved = false; // center on the PC?

#define INF 999999

void initcell(cell *c) {
  c->mpdist = INF;   // minimum distance from the player, ever
  c->cpdist = INF;   // current distance from the player
  c->pathdist = INF; // current distance from the player, along paths (used by yetis)
  c->heat = 0; c->iswall = false; c->item = 0; 
  c->monst = 0;      // yeti=1 wolf=2
  }

cellwalker cwt; // player character position

// reduce c->mpdist to d; also generate the landscape

void setdist(cell *c, int d) {
  if(c->mpdist <= d) return;
  c->mpdist = d;
  if(d < 10) {
    explore[d]++;
    for(int i=0; i<c->type; i++) {
      createMov(c, i);
      setdist(c->mov[i], d+1);
      }
    
    if(d == 9 && rand() % 100 < 5) {
      c->iswall = true;
      for(int i=0; i<c->type; i++) if(rand() % 100 < 50) {
        createMov(c, i);
        cell *c2 = c->mov[i];
        c2->iswall = true;
        for(int j=0; j<c2->type; j++) if(rand() % 100 < 20) {
          createMov(c2, j);
          cell *c3 = c2->mov[j];
          c3->iswall = true;
          }
        }
      }
    
    if(d == 7 && !c->iswall) {
      if(rand() % 5000 < 100 + 2 * (kills[1] + kills[2])) c->item = 1;
      else if(rand() % 5000 < gold) c->monst = 1 + rand() % 2;
      }
    }
  }

vector<cell*> dcal;  // queue for cpdist
vector<cell*> pathq; // queue for pathdist

bool cblocked(cell *c) {
  return c->iswall || c->monst;
  }

// calculate cpdist and pathdist
void bfs() {
  
  for(int i=0; i<size(dcal); i++) dcal[i]->cpdist = INF;
  dcal.clear();
  
  dcal.push_back(cwt.c);
  cwt.c->cpdist = 0;
  int qb = 0;
  while(qb < size(dcal)) {
    cell *c = dcal[qb++];
    int d = c->cpdist;
    if(d == 9) break;
    for(int i=0; i<c->type; i++) {
      // printf("i=%d cd=%d\n", i, c->mov[i]->cpdist);
      if(c->mov[i]->cpdist > d+1) {
        c->mov[i]->cpdist = d+1;
        dcal.push_back(c->mov[i]);
        }
      }
    }

  for(int i=0; i<size(pathq); i++) pathq[i]->pathdist = INF;
  pathq.clear();
  
  pathq.push_back(cwt.c);
  cwt.c->pathdist = 0;
  qb = 0;
  while(qb < size(pathq)) {
    cell *c = pathq[qb++];
    if(c->cpdist > 7) continue;
    int d = c->pathdist;
    if(cblocked(c)) continue;
    for(int i=0; i<c->type; i++) {
      // printf("i=%d cd=%d\n", i, c->mov[i]->cpdist);
      if(c->mov[i]->pathdist == INF) {
        c->mov[i]->pathdist = d+1;
        pathq.push_back(c->mov[i]);
        }
      }
    }
  }

// move monsters and heat
void heat() {
  int oldmelt = kills[0];
  for(int i=0; i<size(pathq); i++) {
    cell *c = pathq[i];
    if(c->monst > 0) {
      if(c->pathdist == 1) {
        // c->iswall = true; c->ismon = false;
        playerdead = true;
        break;
        }
      
      if(c->monst == 1) {
        int nc = 0;
        for(int j=0; j<c->type; j++) if(c->mov[j]->pathdist < c->pathdist && !cblocked(c->mov[j])) nc++;
        if(nc == 0) {
          continue;
          }
        nc = rand() % nc;
        for(int j=0; j<c->type; j++) if(c->mov[j]->pathdist < c->pathdist && !cblocked(c->mov[j])) {
          if(!nc) {
            c->mov[j]->monst = c->monst, c->monst = 0;
            break;
            }
          nc--;
          }
        }
      
      else if(c->monst == 2) {
        int bhd = -1;
        ld besth = c->heat;
        for(int j=0; j<c->type; j++) if(c->mov[j]->heat > besth && !cblocked(c->mov[j]))
          besth = c->mov[j]->heat, bhd = j;
        if(bhd >= 0) {
          // printf("wolf moved from %Lf (%p) to %Lf (%p)\n", c->heat, c, besth, c->mov[bhd]);
          c->mov[bhd]->monst = -c->monst, c->monst = 0;
          }
        }
      }
    }

  for(int i=0; i<size(pathq); i++) {
    cell *c = pathq[i];
    if(c->monst < 0) c->monst = -c->monst;
    }

  cwt.c->heat += 1;
  for(int i=0; i<size(dcal); i++) {
    cell *c = dcal[i];
    if(c->cpdist > 8) break;
    
    ld hmod = 0;
    
    for(int j=0; j<c->type; j++) {
      ld hdiff = c->mov[j]->heat - c->heat;
      hdiff /= 10;
      hmod += hdiff;
      c->mov[j]->heat -= hdiff;
      }
    
    c->heat += hmod;
    if(c->iswall & c->heat > .4) c->iswall = false, kills[0]++;
    }
  
  if(kills[0] != oldmelt) bfs();
  }

// how many monsters are near
int monstersnear(cell *c) {
  int res = 0;
  for(int t=0; t<c->type; t++) if(c->mov[t]->monst) res++;
  return res;
  }

// move the PC in direction d (or stay in place for d == -1)
void movepcto(int d) {
  playermoved = true;
  if(d >= 0) {
    cell *c2 = cwt.c->mov[d];
    if(c2->iswall)
      return;
    else if(c2->monst) {
      if(monstersnear(cwt.c) > 1) {
        printf("You would be killed by another monster!\n");
        return;
        }
      kills[c2->monst]++; 
      c2->monst = 0;
      bfs();
      }
    else {
      if(monstersnear(c2)) {
        printf("A monster would kill you there!\n");
        return;
        }
      if(c2->item) c2->item = 0, gold++;
      cwt.spin = d; cwstep(cwt);
      setdist(cwt.c, 0);
      bfs();
      }
    }
  if(monstersnear(cwt.c) > 0) {
    printf("A monster would get you!\n");
    return;
    }
  heat();
  }

// initialize the game
void initgame() {
  cwt.c = origin.c7; cwt.spin = 0;
  for(int i=9; i>=0; i--)
    setdist(cwt.c, i);
  cwt.c->iswall = false;
  cwt.c->item = 0;
  cwt.c->monst = 0;
  bfs();
  }

