// Hyperbolic RogueH:\proj\progs\rogue\hyper\graph.cpp
// Copyright (C) 2011 Zeno Rogue, see 'hyper.cpp' for details

// basic graphics:

SDL_Surface *s;
TTF_Font *font[256];

#define COLORGOLD 0xFFD700
#define COLORYETI 0x4040FF
#define COLORWOLF 0x8040D0

// x resolutions

#define NUMMODES 5

int xres[NUMMODES] = { 640, 800, 1024, 1280, 1600 };
int yres[NUMMODES] = { 480, 600,  768, 1024, 1200 };

// get coordinates of H in the Poincare model
// change bz to 0 to get Klein model

struct videopar {
  int modeid;
  ld scale, eye, alpha, aspeed;
  bool full;
  bool goteyes;
  bool quick;
  
  int xres, yres;
  
  // paramaters calculated from the above
  int xcenter, ycenter;
  int radius;
  ld alphax, beta;
  };

videopar vid;
 
int mousex, mousey, mousedist, mousedest;

int ZZ;

int& qpixel(SDL_Surface *surf, int x, int y) {
  if(x<0 || y<0 || x >= surf->w || y >= surf->h) return ZZ;
  char *p = (char*) surf->pixels;
  p += y * surf->pitch;
  int *pi = (int*) (p);
  return pi[x];
  }

void displaystr(int x, int y, int shift, int size, const char *str, int color, int align) {

  if(size <= 0 || size > 255) {
    printf("size = %d\n", size);
    return;
    }
  SDL_Color col;
  col.r = (color >> 16) & 255;
  col.g = (color >> 8 ) & 255;
  col.b = (color >> 0 ) & 255;

  if(!font[size])
    font[size] = TTF_OpenFont("VeraBd.ttf", size);

  SDL_Surface *txt = TTF_RenderText_Solid(font[size], str, col);
  
  SDL_Rect rect;

  rect.w = txt->w;
  rect.h = txt->h;

  rect.x = x - rect.w * align / 16;
  rect.y = y - rect.h/2;
  
  if(shift) {
    SDL_Surface* txt2 = SDL_DisplayFormat(txt);
    SDL_LockSurface(txt2);
    int c0 = qpixel(txt2, 0, 0);
    for(int yy=0; yy<rect.h; yy++)
    for(int xx=0; xx<rect.w; xx++) if(qpixel(txt2, xx, yy) != c0)
      qpixel(s, rect.x+xx-shift, rect.y+yy) |= color & 0xFF0000,
      qpixel(s, rect.x+xx+shift, rect.y+yy) |= color & 0x00FFFF;
    SDL_UnlockSurface(txt2);
    SDL_FreeSurface(txt2);
    }
  else {
    SDL_UnlockSurface(s);
    SDL_BlitSurface(txt, NULL, s,&rect); 
    SDL_LockSurface(s);
    }
  SDL_FreeSurface(txt);
  }

void displaychr(int x, int y, int shift, int size, char chr, int col) {

  char buf[2];
  buf[0] = chr; buf[1] = 0;
  displaystr(x, y, shift, size, buf, col, 8);
  }

void displaynum(int x, int y, int shift, int size, int col, int val, const char *title) {
  char buf[64];
  sprintf(buf, "%d", val);
  displaystr(x-8, y, shift, size, buf, col, 16);
  displaystr(x, y, shift, size, title, col, 0);
  }

void displayfloat(int x, int y, int shift, int size, int col, const char *format, ld val, const char *title) {
  char buf[64];
  sprintf(buf, format, val);
  displaystr(x-8, y, shift, size, buf, col, 16);
  displaystr(x, y, shift, size, title, col, 0);
  }

void getcoord(const hyperpoint& H, int& x, int& y, int &shift) {
  
  if(H[2] < 0.999) {
    printf("error: %s\n", display(H));
    exit(1);
    }
  ld tz = vid.alphax+H[2];
  x = vid.xcenter + int(vid.radius * H[0] / tz);
  y = vid.ycenter + int(vid.radius * H[1] / tz);
  shift = vid.goteyes ? int(vid.eye * vid.radius * (1 - vid.beta / tz)) : 0;
  }

void drawline(const hyperpoint& H1, int x1, int y1, const hyperpoint& H2, int x2, int y2, int col) {
  if((x1-x2)*(x1-x2)+(y1-y2)*(y1-y2) <= 1) return;
  
  hyperpoint H3 = mid(H1, H2);
  int x3, y3, s3; getcoord(H3, x3, y3, s3);
  qpixel(s, x3-s3, y3) |= col & 0xFF0000;
  qpixel(s, x3+s3, y3) |= col & 0x00FFFF;

  drawline(H1, x1, y1, H3, x3, y3, col);
  drawline(H3, x3, y3, H2, x2, y2, col);
  
  }

void drawline(const hyperpoint& H1, const hyperpoint& H2, int col) {
  int x1, y1, s1; getcoord(H1, x1, y1, s1);
  int x2, y2, s2; getcoord(H2, x2, y2, s2);
  drawline(H1, x1, y1, H2, x2, y2, col);
  }

// game-related graphics

transmatrix View; // current rotation, relative to viewctr
transmatrix cwtV; // player-relative view
heptspin viewctr; // heptagon and rotation where the view is centered at
bool playerfound; // has player been found in the last drawing?

#include "geometry.cpp"

void drawcell(cell *c, transmatrix V) {
  
  // draw a web-like map
  if(0) if(c->type == 6) {
    for(int a=0; a<3; a++)
    drawline(V*Crad[a*7], V*Crad[a*7+21], 0xd0d0);
    }
  else {
    for(int a=0; a<7; a++)
    drawline(V*C0, V*Crad[(21+a*6)%42], 0xd0d0);
    }
  
  // save the player's view center
  if(c == cwt.c) {
    playerfound = true;
    cwtV = V * spin(-cwt.spin * 2*M_PI/c->type) * spin(M_PI);
    }

  if(1) {
    int xc, yc, sc, xs, ys, ss;
    getcoord(V*C0, xc, yc, sc);
    getcoord(V*xpush(.5)*C0, xs, ys, ss);
    // int col = 0xFFFFFF - 0x20 * c->maxdist - 0x2000 * c->cpdist;
    int col = 0xFFFFFF - 0x101 * (min(int(c->heat*600), 255));
    char ch = '.'; // '0' + c->pathdist;
    if(c->mpdist > 8) return; // not yet generated
    else if(c->monst == 1) ch = 'Y', col = COLORYETI;
    else if(c->monst == 2) ch = 'w', col = COLORWOLF;
    else if(c->item) ch = '$', col = COLORGOLD;
    else if(c->iswall) ch = '#';
    else col = (col & 0xFEFEFE) / 2;
    if(c->cpdist == 0) { ch = '@'; col *= 2; }
    displaychr(xc, yc, sc, int(sqrt(squar(xc-xs)+squar(yc-ys))), ch, col);
    
    // process mouse
    
    for(int i=-1; i<cwt.c->type; i++) if(i == -1 ? cwt.c == c : cwt.c->mov[i % cwt.c->type] == c) {
      int ndist = (xc-mousex) * (xc-mousex) + (yc-mousey) * (yc-mousey);
      if(ndist < mousedist) mousedist = ndist, mousedest = i;
      }
    }
  }

void drawrec(const heptspin& hs, int lev, hstate s, transmatrix V) {
  
  drawcell(hs.h->c7, V * spin(hs.spin*2*M_PI/7));
  
  if(lev == 0) return;
  
  for(int d=0; d<7; d++) {
    int ds = fixrot(hs.spin + d);
    createMov(hs.h->c7, ds);
    if(hs.h->c7->spn[ds] == 0)
      drawcell(hs.h->c7->mov[ds], V * hexmove[d]);
    }

  for(int d=0; d<7; d++) {
    hstate s2 = transition(s, d);
    if(s2 == hsError) continue;
    heptspin hs2 = hsstep(hsspin(hs, d), 0);
    drawrec(hs2, lev-1, s2, V * heptmove[d]);
    }
  
  }

void drawthemap() {
  mousedist = INF;
  mousedest = -1;
  playerfound = false;
  drawrec(viewctr, 4, hsOrigin, View);
  }

void centerpc(ld aspd) { 
  hyperpoint H = cwtV * C0;
  ld R = sqrt(H[0] * H[0] + H[1] * H[1]);
  if(R < 1e-9) return;
  aspd *= (1+R);
  if(R < aspd) {
    hyperpoint H2 = spintox(H) * H;
    View = rspintox(H) * pushxto0(H2) * spintox(H) * View;
    }
  else 
    View = rspintox(H) * xpush(-aspd) * spintox(H) * View;
  }

void drawmovestar() {

  if(!playerfound) return;

  hyperpoint H = cwtV * C0;
  ld R = sqrt(H[0] * H[0] + H[1] * H[1]);
  transmatrix Centered = Id;
  if(R > 1e-9) {
    hyperpoint H2 = spintox(H) * H;
    Centered = rspintox(H) * rpushxto0(H2) * spintox(H) * Centered;
    }
  
  for(int d=0; d<8; d++) 
    drawline(Centered * C0, Centered * spin(M_PI*d/4)* xpush(.5) * C0, vid.goteyes? 0xE08060 : 0xC00000);
  }

void optimizeview() {
  
  int turn;
  ld best = INF;
  
  transmatrix TB;
  
  for(int i=-1; i<7; i++) {

    ld trot = -i * M_PI * 2 / 7.0;
    transmatrix T = i < 0 ? Id : spin(trot) * xpush(tessf) * spin(M_PI);
    hyperpoint H = View * T * C0;
    if(H[2] < best) best = H[2], turn = i, TB = T;
    }
  
  if(turn >= 0) {
    View = View * TB;
    fixmatrix(View);
    viewctr = hsspin(viewctr, turn);
    viewctr = hsstep(viewctr, 0);
    }
  }

void movepckeydir(int d) {
  hyperpoint H = cwtV * C0;
  ld R = sqrt(H[0] * H[0] + H[1] * H[1]);
  transmatrix Centered = cwtV;
  if(R > 1e-9) {
    hyperpoint H2 = spintox(H) * H;
    Centered = rspintox(H) * pushxto0(H2) * spintox(H) * Centered;
    }
  int bdir = -1;
  ld binv = 99;
  hyperpoint MT = spin(-d * M_PI/4) * xpush(1) * C0;
  for(int i=0; i<cwt.c->type; i++) {
    ld inv = intval(Centered * spin(-i * 2 * M_PI /cwt.c->type) * xpush(1) * C0, MT);
    if(inv < binv) binv = inv, bdir = i;
    }
  cwspin(cwt, bdir);
  movepcto(cwt.spin);
  }

void calcparam() {
  vid.xcenter = vid.xres / 2;
  vid.ycenter = vid.yres / 2;
  vid.radius = int(vid.scale * vid.ycenter);
  vid.beta = 1 + vid.alpha + vid.eye;
  vid.alphax = vid.alpha + vid.eye;
  vid.goteyes = vid.eye > 0.001 || vid.eye < -0.001;
  }

int paruntil;

void showpar() {
  paruntil = SDL_GetTicks() + 500;
  }

void drawscreen() {
  calcparam();
  SDL_LockSurface(s);
  // unsigned char *b = (unsigned char*) s->pixels;
  // int n = vid.xres * vid.yres * 4;
  // while(n) *b >>= 1, b++, n--;
  memset(s->pixels, 0, vid.xres * vid.yres * 4);
  
  if(!vid.goteyes) for(int r=0; r<1500; r++)
    qpixel(s, vid.xcenter + int(vid.radius * sin(r)), vid.ycenter + int(vid.radius * cos(r))) = 0xFF;
    
  if(1) for(int t=0; t<int(lines.size()); t++) drawline(View * lines[t].P1, View * lines[t].P2, lines[t].col);

  drawthemap();
  drawmovestar();
  
  char buf[16];
  int vx = vid.xres - 60;
  displaynum(vx, 25, 0, 24, COLORGOLD, gold, "$");
  displaynum(vx, 50, 0, 24, COLORYETI, kills[1], "Y");
  displaynum(vx, 75, 0, 24, COLORWOLF, kills[2], "w");
  displaynum(vx,100, 0, 24, 0xc0c0c0, kills[2], ":");
  displaynum(vx,125, 0, 24, 0xFFFFFF, kills[0], "#");
  
  if(SDL_GetTicks() < paruntil) {
    int vy = vid.yres;
    displayfloat(vx, vy-100, 0, 24, 0xFFFFFF, "%4.2Lf", vid.alpha,  "p");
    displayfloat(vx, vy- 75, 0, 24, 0xFFFFFF, "%4.2Lf", vid.scale,  "z");
    displayfloat(vx, vy- 50, 0, 24, 0xFFFFFF, "%4.2Lf", vid.eye*10, "e");
    displayfloat(vx, vy- 25, 0, 24, 0xFFFFFF, "%4.2Lf", vid.aspeed, "a");
    }
  
  SDL_UnlockSurface(s);
  
  SDL_UpdateRect(s, 0, 0, vid.xres, vid.yres);
  }

void setvideomode() {
  if(vid.modeid < 0) vid.modeid = 0;
  if(vid.modeid >= NUMMODES) vid.modeid = NUMMODES - 1;
  vid.xres = xres[vid.modeid];
  vid.yres = yres[vid.modeid];
  s= SDL_SetVideoMode(vid.xres, vid.yres, 32, vid.full ? SDL_FULLSCREEN : 0);
  }

void initgraph() {

  vid.modeid = 2;
  vid.scale = 1;
  vid.alpha = 1;
  vid.aspeed = 0;
  vid.eye = 0;
  vid.full = false;
  vid.quick = true;
  
  viewctr.h = &origin;
  viewctr.spin = 0;
  View = Id;

  initgeo();

  SDL_Init(SDL_INIT_VIDEO);
  setvideomode();
  if(!s) {
    printf("Failed to initialize graphics.\n");
    exit(2);
    }
    
  SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY, SDL_DEFAULT_REPEAT_INTERVAL);
  
  if(TTF_Init() != 0) {
    printf("Failed to initialize TTF.\n");
    exit(2);
    }

  }

int frames;

void mainloop() {
  int curt, lastt;
  curt = SDL_GetTicks();
  while(true) {
    frames++;
    optimizeview();
    lastt = curt; curt = SDL_GetTicks();
    if(playermoved && vid.aspeed > -5.01)
      centerpc((curt - lastt) / 1000.0 * exp(vid.aspeed));
    drawscreen();
    SDL_Event ev;
    while(SDL_PollEvent(&ev)) {
      if(ev.type == SDL_KEYDOWN) {
        int sym = ev.key.keysym.sym;
        
        ld shift = 1;
        if(ev.key.keysym.mod & (KMOD_LSHIFT | KMOD_RSHIFT)) shift = -1;
        if(ev.key.keysym.mod & (KMOD_LCTRL | KMOD_RCTRL)) shift /= 10;

        if(sym == SDLK_RIGHT) View = xpush(-0.2*shift) * View, playermoved = false;
        if(sym == SDLK_LEFT) View = xpush(+0.2*shift) * View, playermoved = false;
        if(sym == SDLK_UP) View = ypush(+0.2*shift) * View, playermoved = false;
        if(sym == SDLK_DOWN) View = ypush(-0.2*shift) * View, playermoved = false;
        if(sym == SDLK_PAGEUP) View = spin(0.2*shift) * View;
        if(sym == SDLK_PAGEDOWN) View = spin(-0.2*shift) * View;
        
        if(sym == SDLK_KP6) movepckeydir(0);
        if(sym == SDLK_KP3) movepckeydir(1);
        if(sym == SDLK_KP2) movepckeydir(2);
        if(sym == SDLK_KP1) movepckeydir(3);
        if(sym == SDLK_KP4) movepckeydir(4);
        if(sym == SDLK_KP7) movepckeydir(5);
        if(sym == SDLK_KP8) movepckeydir(6);
        if(sym == SDLK_KP9) movepckeydir(7);
        
        if(sym == 'l') movepckeydir(0);
        if(sym == 'n') movepckeydir(1);
        if(sym == 'j') movepckeydir(2);
        if(sym == 'b') movepckeydir(3);
        if(sym == 'h') movepckeydir(4);
        if(sym == 'y') movepckeydir(5);
        if(sym == 'k') movepckeydir(6);
        if(sym == 'u') movepckeydir(7);

        if(sym == '.') movepcto(-1);
        if(sym == SDLK_KP5) movepcto(-1);

        if(sym == 'q') return;

        if(sym == 'c' || sym == 'r' || sym == SDLK_HOME) 
          if(playerfound) centerpc(INF);
          else {
            View = Id;
            viewctr.h = cwt.c->master;
            SDL_LockSurface(s);
            drawthemap(); 
            SDL_UnlockSurface(s);
            centerpc(INF);
            }
        
        if(sym == 'p') showpar(), vid.alpha += shift * 0.1;
        
        if(sym == 'e') showpar(), vid.eye += shift * 0.01;
        
        if(sym == 'z') showpar(), vid.scale += shift * 0.1;

        if(sym == 'a') showpar(), vid.aspeed += shift;

        if(sym == 'x') showpar(), vid.modeid += (shift > 0 ? 1 : -1), setvideomode();
        
        if(sym == 'f') showpar(), vid.full = !vid.full, setvideomode();
        
        // if(sym == 'y') cwspin(cwt, -1);
        // if(sym == 'u') cwspin(cwt, +1);
        // if(sym == 'i') cwstep(cwt);

        }
      if(ev.type == SDL_MOUSEMOTION) {
        mousex = ev.motion.x;
        mousey = ev.motion.y;
        }
      if(ev.type == SDL_MOUSEBUTTONDOWN) {
        movepcto(mousedest);
        }
      if(ev.type == SDL_QUIT)
        return;
      }
    
    if(playerdead) break;
    }
  
  }
