// Hyperbolic Rogue
// Copyright (C) 2011 Zeno Rogue

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

#include <SDL/SDL.h>
#undef main

#include <SDL/SDL_ttf.h>
#include <math.h>
#include <time.h>
#include <vector>
using namespace std;

typedef long double ld;

template<class T> int size(T x) {return x.size(); }

#include "hyperpoint.cpp"
#include "heptagon.cpp"
#include "cell.cpp"
#include "game.cpp"
#include "graph.cpp"

int main() {

  printf("Hyperbolic Rogue v1.0 by Zeno Rogue <zeno@attnam.com>\n");
  printf("released under GNU General Public License version 2 and thus\n");
  printf("comes with absolutely no warranty; see COPYING for details\n");

  srand(time(NULL));
  
  initcells();
  initgame();
  initgraph();
  
  int t1 = SDL_GetTicks();
  
  mainloop();
  SDL_Quit();

  int msec = SDL_GetTicks() - t1;
  printf("frame : %f ms (%f fps)\n", 1.*msec/frames, 1000.*frames/msec);
  if(playerdead) {
    printf("You have been killed by a monster!\n");
    }
  printf("cells generated: %d\n", cellcount);
  printf("Number of cells explored, by distance from the player:\n"); for(int i=0; i<10; i++) printf(" %d", explore[i]); printf("\n");
  printf("gold collected: %d\n", gold);
  printf("walls melted: %d\n", kills[0]);
  printf("yetis killed: %d\n", kills[1]);
  printf("wolves killed: %d\n", kills[2]);
  printf("heptagons travelled: %d\n", celldist(cwt.c));
  return 0;
  }
